"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.createDatabase = exports.SyncDatabase = exports.Database = exports.DatabaseObject = exports.DatabaseArray = exports.getNormalValue = exports.createDatabaseValue = void 0;
function createDatabaseValue(val) {
    if (val instanceof DatabaseObject || val instanceof DatabaseArray)
        return val;
    if (val instanceof Array)
        return new DatabaseArray(val);
    if (val instanceof Object)
        return new DatabaseObject(val);
    return val;
}
exports.createDatabaseValue = createDatabaseValue;
function getNormalValue(val) {
    if (val instanceof DatabaseObject || val instanceof DatabaseArray)
        return val.data;
    else
        return val;
}
exports.getNormalValue = getNormalValue;
class DatabaseArray {
    constructor(data = []) { this.data = data; }
    set data(data) {
        this._data = [];
        data.forEach((v, k) => this._data[k] = createDatabaseValue(v));
    }
    get data() {
        let data = [];
        for (let i = 0; i < this._data.length; i++)
            data[i] = getNormalValue(this._data[i]);
        return data;
    }
    get length() {
        return this._data.length;
    }
    set(key, value) {
        if (typeof key == "string")
            this.set(key.split(/[.\[\]]/g), value);
        else if (typeof key == "number")
            this._data[key] = createDatabaseValue(value);
        else {
            if (key.length > 1) {
                if (this._data[key[0]] == null)
                    this._data[key[0]] = createObjectForKey(key);
                let e = this._data[key[0]];
                if (!(e instanceof DatabaseObject || e instanceof DatabaseArray))
                    throw new Error(`Can't use set property of ${typeof e}`);
                e.set(key.slice(1, key.length), value);
            }
            else {
                this._data[typeof key[0] == "string" ? parseInt(key[0]) : key[0]] = createDatabaseValue(value);
            }
        }
        return this;
    }
    setDefault(key, value) {
        if (!this.contains(key))
            this.set(key, value);
        return this;
    }
    get(key) {
        if (typeof key == "string")
            return this.get(key.split(/[.\[\]]/g));
        else if (typeof key == "number")
            return this._data[key];
        else {
            if (key.length > 1) {
                let e = this._data[key[0]];
                if (!(e instanceof DatabaseObject || e instanceof DatabaseArray))
                    return undefined;
                return e === null || e === void 0 ? void 0 : e.get(key.slice(1, key.length));
            }
            else {
                return this._data[typeof key[0] == "string" ? parseInt(key[0]) : key[0]];
            }
        }
    }
    getNormal(key) {
        return getNormalValue(this.get(key));
    }
    update(key, update) {
        this.set(key, update(this.get(key)));
        return this;
    }
    contains(key) {
        if (typeof key == "string")
            return this.contains(key.split(/[.\[\]]/g));
        else if (typeof key == "number")
            return this._data[key] != null;
        else {
            if (key.length > 1) {
                let e = this._data[key[0]];
                if (!(e instanceof DatabaseObject || e instanceof DatabaseArray))
                    return false;
                return e === null || e === void 0 ? void 0 : e.contains(key.slice(1, key.length));
            }
            else {
                return (this._data[typeof key[0] == "string" ? parseInt(key[0]) : key[0]]) != null;
            }
        }
    }
    push(...values) {
        this._data.push(...values.map(v => createDatabaseValue(v)));
        return this;
    }
}
exports.DatabaseArray = DatabaseArray;
class DatabaseObject {
    constructor(data = {}) { this.data = data; }
    set data(data) {
        this._data = [];
        Object.keys(data).forEach(k => this._data[k] = createDatabaseValue(data[k]));
    }
    get data() {
        let data = {};
        Object.keys(this._data).forEach(k => data[k] = getNormalValue(this._data[k]));
        return data;
    }
    get length() {
        return Object.keys(this._data).length;
    }
    set(key, value) {
        if (typeof key == "string")
            return this.set(key.split(/[.\[\]]/g), value);
        else {
            if (key.length > 1) {
                if (this._data[key[0]] == null)
                    this._data[key[0]] = createObjectForKey(key);
                let e = this._data[key[0]];
                if (!(e instanceof DatabaseObject || e instanceof DatabaseArray))
                    throw new Error(`Can't use set property of ${typeof e}`);
                e.set(key.slice(1, key.length), value);
            }
            else {
                this._data[key[0]] = createDatabaseValue(value);
            }
        }
        return this;
    }
    setDefault(key, value) {
        if (!this.contains(key))
            this.set(key, value);
        return this;
    }
    setDefaults(defaults) {
        Object.keys(defaults).forEach(k => {
            if (!this._data[k])
                this._data[k] = createDatabaseValue(defaults[k]);
            else if (defaults[k] instanceof Object && this._data[k] instanceof DatabaseObject)
                this._data[k].setDefaults(defaults[k]);
        });
        return this;
    }
    get(key) {
        if (typeof key == "string")
            return this.get(key.split(/[.\[\]]/g));
        else {
            if (key.length > 1) {
                let e = this._data[key[0]];
                if (!(e instanceof DatabaseObject || e instanceof DatabaseArray))
                    return undefined;
                return e === null || e === void 0 ? void 0 : e.get(key.slice(1, key.length));
            }
            else {
                return this._data[key[0]];
            }
        }
    }
    getNormal(key) {
        return getNormalValue(this.get(key));
    }
    update(key, update) {
        this.set(key, update(this.get(key)));
        return this;
    }
    contains(key) {
        if (typeof key == "string")
            return this.contains(key.split(/[.\[\]]/g));
        else {
            if (key.length > 1) {
                let e = this._data[key[0]];
                if (!(e instanceof DatabaseObject || e instanceof DatabaseArray))
                    return false;
                return e === null || e === void 0 ? void 0 : e.contains(key.slice(1, key.length));
            }
            else {
                return this._data[key[0]] != null;
            }
        }
    }
    generateId(name) {
        let counter = `id_counters.${name}`;
        if (this.get(counter) == null)
            this.set(counter, -1);
        return this.update(counter, (e) => e + 1).get(counter);
    }
}
exports.DatabaseObject = DatabaseObject;
class Database extends DatabaseObject {
    constructor(adapter, data) {
        super(data);
        this.adapter = adapter;
        this.adapter = adapter;
    }
    saveData() {
        return __awaiter(this, void 0, void 0, function* () { yield this.adapter.save(this.data); return this; });
    }
    reloadData() {
        return __awaiter(this, void 0, void 0, function* () { this.data = yield this.adapter.load(); return this; });
    }
}
exports.Database = Database;
class SyncDatabase extends DatabaseObject {
    constructor(adapter, data) {
        super(data || adapter.load());
        this.adapter = adapter;
        this.adapter = adapter;
    }
    saveData() { this.adapter.save(this.data); return this; }
    reloadData() { this.data = this.adapter.load(); return this; }
}
exports.SyncDatabase = SyncDatabase;
function createDatabase(adapter) {
    let result = adapter.load();
    if (result instanceof Promise) {
        return new Promise((rs) => {
            result.then((data) => {
                rs(new Database(adapter, data));
            });
        });
    }
    return new SyncDatabase(adapter, result);
}
exports.createDatabase = createDatabase;
function createObjectForKey(key) {
    if (typeof key == "string")
        return createObjectForKey(key.split(/[.\[\]]/g));
    if (typeof key[1] == 'number' || (typeof key[1] == 'string' && isNumberString(key[1])))
        return new DatabaseArray([]);
    return new DatabaseObject();
}
const numberCharacters = "0123456789";
function isNumberString(s) {
    for (let i = 0; i < s.length; i++)
        if (!numberCharacters.includes(s.charAt(i)))
            return false;
    return true;
}
;
